/* wptGPGME.cpp - WinPT GPGME interface
 *	Copyright (C) 2001-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <windows.h>

#include "resource.h"
#include "wptNLS.h"
#include "wptGPG.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptRegistry.h"
#include "wptDlgs.h"

#include "openpgp.h"

BOOL CALLBACK  keycache_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam);
void progress_cleanup (progress_filter_s * pfx);

static gpg_keycache_t pub = NULL;
static gpg_keycache_t sec = NULL;
static char *gpg_secring = NULL;


/* Reload the key cache. */
void
keycache_reload (HWND dlg) 
{	
    refresh_cache_s rcs;

    memset (&rcs, 0, sizeof rcs);
    rcs.kr_reload = rcs.kr_update = 1;
    rcs.tr_update = 0;
    DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_KEYCACHE, dlg,					
		    keycache_dlg_proc, (LPARAM)&rcs);
}


/* Release both key cache objects. If @cleanup is 1,
   also release other global structs. */
void
keycache_release (int cleanup)
{
    int n = gpg_keycache_get_size (pub);
    char tmpbuf[64];

    /* XXX: update the value when the cache has changed. */
    sprintf (tmpbuf, "%d", n);
    set_reg_key (HKEY_CURRENT_USER, "Software\\WinPT", "nKeys", tmpbuf);

    if (pub) {
        gpg_keycache_release (pub); 
        pub = NULL;
    }
    if (sec) {
        gpg_keycache_release (sec);
        sec = NULL;
    }
    if (cleanup) {
	if (gpg_secring)
	    free (gpg_secring);
	gpg_secring = NULL;
    }
}


/* Update the key with the keyid @keyid in the key cache.
   If @is_sec is 1, the secret key cache is used. */
gpgme_error_t
keycache_update (int is_sec, const char *keyid)
{
    gpg_keycache_t ctx = pub;
    gpgme_error_t err;

    if (is_sec)
	ctx = sec;
    err = gpg_keycache_update_key (ctx, is_sec, pub, keyid);
    if (is_sec)
	gpg_keycache_prepare_single (ctx, keyid, NULL, gpg_secring);
    return err;
}


/* Initialize both cache contexts. Use @pubring for the public
   keyring and @secring for the secret keyring. */
gpgme_error_t
keycache_init (const char *pubring, const char *secring)
{
    struct progress_filter_s pfx;
    gpgme_error_t err;
    int val = 0;
    char *p;

    if (secring != NULL) {
	free_if_alloc (gpg_secring);
	gpg_secring = get_gnupg_keyring (0, NO_STRICT);
    }
    
    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "nKeys");
    if (p && *p != ' ') {
	val = atoi (p);
	free_if_alloc (p);	
    }

    memset (&pfx, 0, sizeof (pfx));
    /* Release old contexts first. */
    keycache_release (0);

    err = gpg_keycache_new (&pub);
    if (err)
	return err;
    if (val != 0)
	gpg_keycache_set_cb (pub, progress_callback, &pfx, val);
    err = gpg_keycache_new (&sec);
    if (!err)
	err = gpg_keycache_init (pub, NULL, 0);
    if (!err)
	err = gpg_keycache_init (sec, NULL, 1);
    if (!err && pubring && *pubring)
	err = gpg_keycache_prepare (pub, pubring, NULL);
    if (!err && secring && * secring)
	err = gpg_keycache_prepare (sec, NULL, secring);
    if (!err)
	gpg_keycache_sync (pub, sec);
    if (val != 0)
	progress_cleanup (&pfx);
    return err;
}


/* Return the public cache context if @is_pub is set
   the secre cache context otherwise. */
gpg_keycache_t
keycache_get_ctx (int is_pub)
{
    return is_pub? pub : sec;
}


/* Get the GPG key with keyid @keyid from the cache. Return it 
   in @r_key on success. */
static int
get_key_from_cache (const char *keyid, gpgme_key_t *r_key, 
		    struct keycache_s **c, int secret)
{
    gpg_keycache_t cache;
    gpgme_error_t err;
    int mode = secret? KEYCACHE_PRV : KEYCACHE_PUB;

    if (!keyid)
	return WPTERR_GENERAL;
    if (r_key)
	*r_key = NULL;
    cache = keycache_get_ctx (mode);
    if (!cache)
	BUG (0);
    if (!c)
	err = gpg_keycache_find_key (cache, keyid, 0, r_key);
    else
	err = gpg_keycache_find_key2 (cache, keyid, 0, r_key, c);
    return err? WPTERR_GENERAL : 0;
}


/* Get GPG key with keyid @keyid directly from GPG and return
   it in @r_key on success. */
static int
get_key_directly (const char *keyid, gpgme_key_t *r_key, int secret)
{
    gpgme_ctx_t ctx;
    gpgme_error_t err;

    err = gpgme_new (&ctx);
    if (err)
	return WPTERR_GENERAL;
    err = gpgme_get_key (ctx, keyid, r_key, secret);
    gpgme_release (ctx);
    return err? WPTERR_GENERAL : 0;
}


/* Search the public key with @keyid as the keyid in the cache and
   return the item in @k. */
int
winpt_get_pubkey (const char *keyid, winpt_key_s *k)
{
    int rc;
    
    rc = get_key_from_cache (keyid, &k->ctx, &k->ext, 0);
    if (rc)
	return rc;
    k->is_v3 = k->ctx->subkeys->pubkey_algo == GPGME_PK_RSA && 
		strlen (k->ctx->subkeys->fpr) == 32;
    k->is_protected = k->ext->gloflags.is_protected;
    k->keyid = k->ctx->subkeys->keyid;
    k->uid = k->ctx->uids->uid;
    return rc;
}


int
winpt_get_seckey (const char *keyid, winpt_key_s *k)
{
    int rc;
    rc = get_key_from_cache (keyid, &k->ctx, &k->ext, 1);
    if (rc)
	return rc;
    k->is_v3 = k->ctx->subkeys->pubkey_algo == GPGME_PK_RSA && 
		strlen (k->ctx->subkeys->fpr) == 32;
    k->is_protected = k->ext->gloflags.is_protected;
    k->keyid = k->ctx->subkeys->keyid;
    k->uid = k->ctx->uids->uid;
    return rc;
}


int
get_pubkey (const char *keyid, gpgme_key_t *ret_key)
{
    int rc;

    if (pub && sec)
	rc = get_key_from_cache (keyid, ret_key, NULL, 0);
    else
	rc = get_key_directly (keyid, ret_key, 0);
    return rc;
}


int
get_seckey (const char *keyid, gpgme_key_t *ret_skey)
{
    int rc;

    if (pub && sec)
	rc = get_key_from_cache (keyid, ret_skey, NULL, 1);
    else
	rc = get_key_directly (keyid, ret_skey, 1);
    return rc;
}


/* Search for insecure ElGamal keys and return the
   number of founded keys. */
int
count_insecure_elgkeys (void)
{
    gpg_keycache_t pc;
    gpgme_key_t key;
    int n=0;

    pc = keycache_get_ctx (1);
    if (!pc)
	BUG (0);
    while (!gpg_keycache_next_key (pc, 0, &key)) {
	if (key->subkeys->pubkey_algo == GPGME_PK_ELG)
	    n++;
    }
    gpg_keycache_rewind (pc);
    return n;
}



/* Map the signature summary in @sum to signature status table index. 
   Return value: index to table. */
static int
sigsum_to_index (gpgme_sigsum_t sum)
{
    if ((sum & GPGME_SIGSUM_VALID) && (sum & GPGME_SIGSUM_KEY_REVOKED))
	return 7;
    if ((sum & GPGME_SIGSUM_VALID) && (sum & GPGME_SIGSUM_SIG_EXPIRED))
	return 6;
    if (sum & GPGME_SIGSUM_GREEN)
	return 1;
    else if (sum & GPGME_SIGSUM_RED)
	return 2;
    else if (sum & GPGME_SIGSUM_KEY_MISSING)
	return 3;
    return 0;
}


/* Return a humand readable description for the signature status @sum. */
const char*
get_gpg_sigstat (gpgme_sigsum_t sum)
{
    const char *gpg_sigstat[] = {
	_("Error during verification process."),
	_("The signature is good."),
	_("The signature is BAD!"),
	_("The signature could not be checked due to a missing key."),
	_("No valid OpenPGP signature."),
	_("Signature Error"),
	_("Good Signature (Expired Key)"),
	_("Good Signature (Revoked Key)"),
	NULL
    };
    const unsigned int mask = 8;

    return gpg_sigstat[sigsum_to_index (sum) % mask];
}


/* Check if the secret keyring contains at least one 
   key with ultimate trust.
   Return value: 0 on success. */
int
check_ultimate_trusted_key (void)
{
    struct keycache_s *n;

    for (n = sec->item; n; n = n->next) {
	if (n->pubpart && 
	    n->pubpart->key->owner_trust == GPGME_VALIDITY_ULTIMATE)
	return 0;
    }
    return -1;
}
